"""
`ymlparsers` and `parser` modules serves as Low-Level API for this module.

You may need to install some 3-rd party dependencies, in order to use it, you should have installed first. To do it
run `python -m pip install alex-ber-utils[yml]` in order to use it.

See https://medium.com/analytics-vidhya/my-major-init-app-conf-module-1a5d9fb3998c for documentation.
"""
import logging
logger = logging.getLogger(__name__)

from collections import OrderedDict, deque
from pathlib import Path
import io as _io
from alexber.utils.parsers import is_empty, safe_eval as _convert, parse_boolean, ArgumentParser, \
    parse_sys_args as uparse_sys_args
try:
    import alexber.utils.ymlparsers as ymlparsers
except ImportError:
    pass

from .importer import importer


class conf(object):
    """
    See parse_config() function for details.
    """
    GENERAL_KEY = 'general'
    PROFILES_KEY = 'profiles'
    #WHITE_LIST_IMPLICIT_KEY = 'whiteListImplicitely'    #default True
    WHITE_LIST_KEY = 'whiteListSysOverride'
    LIST_ENSURE_KEY ='listEnsure'

    CONFIG_KEY = 'config'
    FILE_LEY = 'file'


class AppConfParser(object):
    def __init__(self, *args, **kwargs):
        self.implicit_convert = kwargs.pop('implicit_convert')
        if self.implicit_convert is None:
            raise ValueError("implicit_convert can't be None")

    basic_parse_sys_args = staticmethod(uparse_sys_args)

    def _parse_yml(self, sys_d, profiles, config_file='config.yml'):
        if ymlparsers.HiYaPyCo.jinja2ctx is None:
            raise ValueError("ymlparsers.HiYaPyCo.jinja2ctx can't be None")

        base_full_path = Path(config_file).resolve()  # relative to cwd

        base_suffix = base_full_path.suffix
        base_name = base_full_path.with_suffix("").name

        with _io.StringIO() as sys_buf:
            ymlparsers.safe_dump(sys_d, sys_buf)

            buffersize = len(profiles) + 1  # default_profile
            yml_files = deque(maxlen=buffersize)

            yml_files.append(str(base_full_path))
            full_path = base_full_path

            for profile in profiles:
                name = base_name + '-' + profile + base_suffix
                full_path = full_path.with_name(name)
                yml_files.append(str(full_path))

            dd = ymlparsers.load([*yml_files, sys_buf.getvalue()])

        return dd

    def _parse_white_list_implicitely(self, default_d):
        subvalue = default_d.get(conf.GENERAL_KEY, {})
        value = subvalue.get(conf.WHITE_LIST_KEY, {})

        ret = is_empty(value)
        if ret is None:
            ret = True
        return ret

    def _parse_white_list(self, default_d):
        calc_implicitely = self._parse_white_list_implicitely(default_d)

        if calc_implicitely:
            white_list = [key for key in default_d.keys() if key not in conf.WHITE_LIST_KEY]
            return white_list

        # default_d[conf.GENERAL_KEY][conf.WHITE_LIST_KEY] is not empty
        subvalue = default_d.get(conf.GENERAL_KEY, {})
        white_list = subvalue.get(conf.WHITE_LIST_KEY, {})

        return white_list

    def _is_white_listed(self, white_list_flat_keys, flat_key):
        b = is_empty(white_list_flat_keys)
        if b:
            return True

        for white_list_flat_key in white_list_flat_keys:
            if flat_key.startswith(white_list_flat_key):
                return True
        return False

    def to_convex_map(self, d, white_list_flat_keys=None):
        dd = OrderedDict()

        for flat_key, value in d.items():
            if not ('.' in flat_key and self._is_white_listed(white_list_flat_keys, flat_key)):
                logger.info(f"Skipping key {flat_key}. It doesn't contain dot.")
                continue

            keys = flat_key.split(".")
            length = len(keys)

            inner_d = dd
            for i, part_key in enumerate(keys):
                if i + 1 == length:  # last element
                    inner_d = inner_d.setdefault(part_key, self.mask_value(value))
                else:
                    inner_d = inner_d.setdefault(part_key, OrderedDict())
        return dd

    def merge_list_value_in_dicts(self, flat_d, d, main_key, sub_key):
        if flat_d is None:
            raise TypeError("flat_d can't be None")

        if d is None:
            raise TypeError("d can't be None")

        flat_key = '.'.join([main_key, sub_key])
        flat_value = self._ensure_list(flat_d, flat_key)

        length_flat_value = len(flat_value)

        subvalue = d.get(main_key, {})
        value = subvalue.get(sub_key, {})

        merged_value = value if length_flat_value == 0 else flat_value
        return merged_value

    def _parse_profiles(self, sys_d, default_d):
        profiles = self.merge_list_value_in_dicts(sys_d, default_d, conf.GENERAL_KEY, conf.PROFILES_KEY)
        b = is_empty(profiles)
        if b:
            profiles = []
        return profiles

    def _parse_list_ensure(self, sys_d, default_d):
        list_ensure = self.merge_list_value_in_dicts(sys_d, default_d, conf.GENERAL_KEY, conf.LIST_ENSURE_KEY)

        b = is_empty(list_ensure)
        if not b:
            # we already proceesed profiles, white_list
            # we should process list_ensure
            profiles_flat_key = '.'.join([conf.GENERAL_KEY, conf.PROFILES_KEY])
            list_ensure_flat_key = '.'.join([conf.GENERAL_KEY, conf.LIST_ENSURE_KEY])
            white_list_flat_key = '.'.join([conf.GENERAL_KEY, conf.WHITE_LIST_KEY])

            list_ensure = [flat_key for flat_key in list_ensure \
                           if flat_key not in {profiles_flat_key, list_ensure_flat_key, white_list_flat_key}]
        else:
            list_ensure = []
        return list_ensure

    def _get_white_listed(self, d, white_list_flat_keys):
        b = is_empty(d)
        if b:
            return d
        dd = OrderedDict()

        for flat_key, value in d.items():
            if not ('.' in flat_key and self._is_white_listed(white_list_flat_keys, flat_key)):
                logger.info(f"Skipping key {flat_key}. It doesn't white listed.")
                continue

            dd[flat_key] = value
        return dd

    def _do_ensure_list(self, flat_d, list_ensure):
        b = is_empty(list)
        if b:
            return

        for flat_key in list_ensure:
            flat_value = self._ensure_list(flat_d, flat_key)
            flat_d[flat_key] = flat_value

    def _bool_is_empty(self, value):
        try:
            ret = parse_boolean(value)
            return ret is None
        except ValueError:
            ret = is_empty(value)
        return ret

    def _bool_convert(self, value):
        try:
            ret = parse_boolean(value)
            return ret
        except ValueError:
            ret = _convert(value)
        return ret

    def mask_value(self, value):
        ret = self._bool_convert(value) \
            if self.implicit_convert \
            else value
        return ret

    def _ensure_list(self, flat_d, key):
        def _ensure_list(v):
            value = str(v)
            elements = value.split(",")

            # empty string will become null
            elements = [None if is_empty(value) else value for value in elements]
            ret = [self.mask_value(value) for value in elements]

            return ret

        if flat_d is None:
            flat_d = {}
        val = flat_d.get(key, None)
        b = self._bool_is_empty(val)
        if b:
            if key in flat_d.keys():
                # we have key:None?
                return [val]
            return []
        return _ensure_list(val)

    def _parse_sys_args(self, argumentParser=None, args=None):
        if ymlparsers.HiYaPyCo.jinja2ctx is None:
            raise ValueError("ymlparsers.HiYaPyCo.jinja2ctx can't be None")

        params, sys_d0 = self.basic_parse_sys_args(argumentParser, args)
        config_file = params.config_file

        full_path = Path(config_file).resolve()  # relative to cwd

        with ymlparsers.DisableVarSubst():
            default_d = ymlparsers.load([str(full_path)])

        white_list = self._parse_white_list(default_d)
        sys_d = self._get_white_listed(sys_d0, white_list)

        profiles = self._parse_profiles(sys_d, default_d)

        list_ensure = self._parse_list_ensure(sys_d, default_d)

        self._do_ensure_list(sys_d, list_ensure)

        dd = self.to_convex_map(sys_d)
        return dd, profiles, white_list, list_ensure, full_path

    def parse_config(self, argumentParser=None, args=None):

        sys_d, profiles, white_list, list_ensure, config_file = self._parse_sys_args(argumentParser=argumentParser,
                                                                                     args=args)
        dd = self._parse_yml(sys_d, profiles, config_file)

        # merge all to dd
        general_d = dd.setdefault(conf.GENERAL_KEY, OrderedDict())
        config_d = general_d.setdefault(conf.CONFIG_KEY, OrderedDict())
        config_d[conf.FILE_LEY] = str(config_file)

        general_d[conf.PROFILES_KEY] = profiles
        general_d[conf.WHITE_LIST_KEY] = white_list
        general_d[conf.LIST_ENSURE_KEY] = list_ensure

        general_d = dd.setdefault(conf.GENERAL_KEY, OrderedDict())
        config_d = general_d.setdefault(conf.CONFIG_KEY, OrderedDict())
        config_d[conf.FILE_LEY] = str(config_file)

        general_d[conf.PROFILES_KEY] = profiles
        general_d[conf.WHITE_LIST_KEY] = white_list
        general_d[conf.LIST_ENSURE_KEY] = list_ensure
        return dd

def _create_default_parser(**kwargs):
    if default_parser_cls is None:
        raise ValueError("default_parser_cls can't be None")

    implicit_convert = kwargs['implicit_convert']
    if implicit_convert is None:
        implicit_convert=default_parser_kwargs['implicit_convert']

    kwargs = {
        **default_parser_kwargs,
        'implicit_convert': implicit_convert,

    }
    parser = default_parser_cls(**kwargs)
    return parser

def mask_value(value, implicit_convert=None):
    """
    If implicit_convert=True, than for system args we assume Python built-in types, including bool,
    and we're converting it to appropriate type.
    Otherwise, implicit_convert will have the value that was set in `intiConfig()`. By default it is True.


    Bool values are case-insensitive.

    :param value: str to ccnvert
    :param implicit_convert: if none, than value that was passed to initConfig() is used (default).
                             if True value attempt to convert value to appropriate type will be done,
                             if False value will be used as is. See mask_value() function.
    :return:
    """
    if default_parser_cls is None:
        raise ValueError("default_parser_cls can't be None")

    parser = _create_default_parser(**{'implicit_convert': implicit_convert})
    ret = parser.mask_value(value)
    return ret


def to_convex_map(d, white_list_flat_keys=None, implicit_convert=None):
    """
    This method receives dictionary with 'flat keys', it has simple key:value structure
    where value can't be another dictionary.
    It will return dictionary of dictionaries with natural key mapping (see bellow),
    optionally, entries will be filtered out according to white_list_flat_keys and,
    optionally, value will be implicitly converted to appropriate type.

    In order to simulate dictionary of dictionaries 'flat keys' compose key from outer dict with key from inner dict
    separated with dot.
    For example, 'general.profiles' 'flat key' corresponds to convex map with 'general' key with dictionary as value
    that have one of the keys 'profiles' with corresponding value.

    if white_list_flat_keys is not None, it will be used to filter out entries from d.
    If implicit_convert=True, than for system args we assume Python built-in types, including bool,
    and we're converting it to appropriate type.
    Otherwise, implicit_convert will have the value that was set in `intiConfig()`. By default it is True.


    :param d: dict with flat keys
    :param white_list_flat_keys: Optional. if present, only keys that start with one of the elements listed here
                                            will be considered.
    :param implicit_convert: if none, than value that was passed to initConfig() is used (default).
                             if True value attempt to convert value to appropriate type will be done,
                             if False value will be used as is. See mask_value() function.
    :return: convex map with optionally filtered entrys
    """
    if default_parser_cls is None:
        raise ValueError("default_parser_cls can't be None")

    parser = _create_default_parser(**{'implicit_convert': implicit_convert})
    dd = parser.to_convex_map(d, white_list_flat_keys)
    return dd



def merge_list_value_in_dicts(flat_d, d, main_key, sub_key, implicit_convert=None):
    """
    This method merge value of 2 dicts. This value represents list of values.

    Value from flat_d is roughly obtained by flat_d[main_key+'.'+sub_key].
    Value from d is roughly obtained by d[main_key][sub_key].

    If value (or intermediate value) is not found empty dict is used.

    This method assumes that flat_d value contain str that represent list (comma-delimited).
    This method assumes that d[main_key] contains dict.
    implicit_convert is applied only for flat_d.


    If implicit_convert=True, than for system args we assume Python built-in types, including bool,
    and we're converting it to appropriate type.
    Otherwise, implicit_convert will have the value that was set in `intiConfig()`. By default it is True.



    :param flat_d: flat dictionary, usually one that was created from parsing system args.
    :param d: dictionary of dictionaries,  usually one that was created from parsing YAML file.
    :param main_key: d[main_key] is absent or dict.
    :param sub_key: d[main_key][sub_key] is absent or list.
    :param implicit_convert: if none, than value that was passed to initConfig() is used (default).
                             if True value attempt to convert value to appropriate type will be done,
                             if False value will be used as is. See mask_value() function.
    :return: merged converted value, typically one from flat_d, if empty than from d
    """
    if default_parser_cls is None:
        raise ValueError("default_parser_cls can't be None")

    parser = _create_default_parser(**{'implicit_convert': implicit_convert})
    merged_value = parser.merge_list_value_in_dicts(flat_d, d, main_key, sub_key)
    return merged_value

def parse_config(argumentParser=None, args=None, implicit_convert=None):
    """
    This is the main function of the module.

    This function parses command line arguments first.
    Than it parse yml files.
    Command line arguments overrides yml files arguments.

    Parameters of yml files we always try to convert on best-effort basses.
    Parameters of system args we try convert according to implicit_convert param (see below).

    In more detail, command line arguments of the form --key=value are parsed first.
    If exists --config_file it's value is used to search for yml file.
    if --config_file is absent, 'config.yml' is used for yml file.
    If yml file is not found, only command line arguments are used.
    If yml file is found, both arguments are used, while
    command line arguments overrides yml arguments.

    --general.profiles or appropriate key in default yml file is used to find 'profiles'.
    Let suppose, that --config_file is resolved to config.yml.
    If 'profiles' is not empty, than it will be used to calculate filenames
    that will be used to override default yml file.
    Let suppose, 'profiles' resolved to ['dev', 'local']. Than first config.yml
    will be loaded, than it will be overridden with config-dev.yml, than
    it will be overridden with config-local.yml.
    At last, it will be overridden with system args.
    This entry can be always be overridden with system args.

    general.whiteListSysOverride key in yml file is optional.
    If not provided, than any key that exists in the default yml file can be overridden
    with system args.
    If provided, than only key that start with one of the key provided here can be used
    to override entrys with system args.
    This entry can't be overridden with system args.

    --general.listEnsure or appropriate key in default yml file is used to instruct that
    listed key should be interpreted as comma-delimited list when is used to override
    entrys with system args.
    This entry can be always be overridden with system args.

    general.config.file is key that is used in returned dict that points to default yml file.

    If implicit_convert=True, than for system args we assume Python built-in types, including bool,
    and we're converting it to appropriate type.
    Otherwise, implicit_convert will have the value that was set in `intiConfig()`. By default it is True.



    :param argumentParser:
    :param args: if not None, suppresses sys.args
    :param implicit_convert: if none, than value that was passed to initConfig() is used (default).
                             if True value attempt to convert value from system args to appropriate type will be done,
                             if False value from system args will be used as is. See mask_value() function.
    :return: dict ready to use
    """
    if default_parser_cls is None:
        raise ValueError("default_parser_cls can't be None")

    if ymlparsers.HiYaPyCo.jinja2ctx is None:
        raise ValueError("ymlparsers.HiYaPyCo.jinja2ctx can't be None")

    parser = _create_default_parser(**{'implicit_convert': implicit_convert})
    dd = parser.parse_config(argumentParser=argumentParser, args=args)
    return dd


default_parser_cls = None
default_parser_kwargs = None

def initConfig(**kwargs):
    """
    This method can be optionally called prior any call to another function in this module.
    It is indented to be called in the MainThread.
    This method can be call with empty params.

    :param default_parser_cls: can be class or str. Optional.
                Default values is: AppConfParser
    :param default_parser_kwargs: this params will be used as default values in default_parser_cls.__init__() function.
                Default values are
                      'implicit_convert':True,
                This means, by default:
                    We're converting values using mask_value() function.

    This method is idempotent.
    :return:
    """
    default_parser_cls_p = kwargs.get('default_parser_cls', None)
    if default_parser_cls_p is None:
        default_parser_cls_p = AppConfParser
    elif isinstance(default_parser_cls_p, str):
        default_parser_cls_p = importer(default_parser_cls_p)
    global default_parser_cls
    default_parser_cls = default_parser_cls_p

    default_parser_kwargs_p = kwargs.get('default_parser_kwargs', {})
    default_parser_kwargs_p = {
        'implicit_convert':True,
        **default_parser_kwargs_p}
    global default_parser_kwargs
    default_parser_kwargs = default_parser_kwargs_p

initConfig()


import functools
import time
from collections import defaultdict, deque
from typing import Any, Optional, Union
from .thread_locals import RLock
from .mains import make_hashable, HashableWrapper
import inspect

class _LRUCache:
    """A simple LRU (Least Recently Used) cache implementation."""

    def __init__(self, maxsize: int):
        self.maxsize = maxsize
        self.cache = {}
        self.order = deque()

    def __getitem__(self, key: Any) -> Any:
        if key not in self.cache:
            raise KeyError(f"Key '{key}' not found in cache.")
        self.order.remove(key)
        self.order.append(key)
        return self.cache[key]

    def __setitem__(self, key: Any, value: Any):
        if key in self.cache:
            self.order.remove(key)
        elif len(self.cache) >= self.maxsize:
            oldest_key = self.order.popleft()
            del self.cache[oldest_key]
        self.cache[key] = value
        self.order.append(key)

    def __len__(self) -> int:
        return len(self.cache)

    def __contains__(self, key: Any) -> bool:
        return key in self.cache
    def clear(self):
        self.cache.clear()
        self.order.clear()


class _LFUCache:
    """A simple LFU (Least Frequently Used) cache implementation."""

    def __init__(self, maxsize: int):
        self.maxsize = maxsize
        self.cache = {}
        self.freq = defaultdict(int)
        self.min_freq = 0
        self.freq_list = defaultdict(deque)

    def __getitem__(self, key: Any) -> Any:
        if key not in self.cache:
            raise KeyError(f"Key '{key}' not found in cache.")
        self._increase_freq(key)
        return self.cache[key]

    def __setitem__(self, key: Any, value: Any):
        if key in self.cache:
            self.cache[key] = value
            self._increase_freq(key)
        else:
            if len(self.cache) >= self.maxsize:
                self._evict()
            self.cache[key] = value
            self.freq[key] = 1
            self.min_freq = 1
            self.freq_list[1].append(key)

    def _increase_freq(self, key: Any):
        freq = self.freq[key]
        self.freq[key] += 1
        self.freq_list[freq].remove(key)
        if not self.freq_list[freq]:
            del self.freq_list[freq]
            if self.min_freq == freq:
                self.min_freq += 1
        self.freq_list[freq + 1].append(key)

    def _evict(self):
        evict_key = self.freq_list[self.min_freq].popleft()
        if not self.freq_list[self.min_freq]:
            del self.freq_list[self.min_freq]
        del self.cache[evict_key]
        del self.freq[evict_key]

    def __len__(self) -> int:
        return len(self.cache)

    def __contains__(self, key: Any) -> bool:
        return key in self.cache

    def clear(self):
        self.cache.clear()
        self.freq.clear()
        self.freq_list.clear()
        self.min_freq = 0


class AsyncCache:
    """
    An asynchronous cache that supports both LFU (Least Frequently Used) and LRU
    (Least Recently Used) eviction policies and an optional Time-to-Live (TTL) for cache entries.
    """

    def __init__(self, maxsize, ttl=None, policy="LFU"):
        """
        Initializes the AsyncCache with the given parameters.

        Args:
            maxsize (int): Maximum size of the cache.
            ttl (Optional[int]): Time-to-Live for cache entries in seconds. Defaults to None.
            policy (str): Cache eviction policy. Can be either "LFU" or "LRU". Defaults to "LFU".
        """
        if policy == "LFU":
            self.cache = _LFUCache(maxsize=maxsize)
        elif policy == "LRU":
            self.cache = _LRUCache(maxsize=maxsize)
        else:
            raise ValueError("Invalid policy. Use 'LFU' or 'LRU'.")

        self.ttl = ttl
        self.expiry_times = {}  # To track expiry times of keys when ttl is not None
        self.lock = RLock()
        self.hits = 0
        self.misses = 0
        self.total_time_ns = 0
        self.total_calls = 0
        self.max_time_ns = 0
        self.min_time_ns = float('inf')

    def _wrap_key(self, key: Any) -> Union[HashableWrapper, Any]:
        try:
            _ = hash(key)
            return key
        except TypeError:
            return HashableWrapper(key)

    async def __getitem__(self, key):
        key = self._wrap_key(key)

        async with self.lock:
            current_time = time.perf_counter_ns()
            if key in self.cache:
                if self.ttl is not None:
                    if current_time < self.expiry_times.get(key, float('inf')):
                        # Cache hit
                        self.hits += 1
                        return self.cache[key]
                    else:
                        # Expired, remove it from both caches
                        self.cache._evict()
                        self.expiry_times.pop(key, None)
                        self.misses += 1
                        raise KeyError(f"Key '{key}' has expired.")
                else:
                    # Cache hit without TTL
                    self.hits += 1
                    return self.cache[key]
            else:
                self.misses += 1
                raise KeyError(f"Key '{key}' not found in cache.")

    async def __setitem__(self, key, value):
        key = self._wrap_key(key)

        async with self.lock:
            self.cache[key] = value
            if self.ttl is not None:
                self.expiry_times[key] = time.perf_counter_ns() + self.ttl * 1e9  # ttl is in seconds, convert to ns

    async def update_profiling(self, exec_time_ns):
        async with self.lock:
            self.total_calls += 1
            self.total_time_ns += exec_time_ns
            self.max_time_ns = max(self.max_time_ns, exec_time_ns)
            self.min_time_ns = min(self.min_time_ns, exec_time_ns)

    async def get_stats(self):
        async with self.lock:
            avg_time_ns = self.total_time_ns / self.total_calls if self.total_calls > 0 else 0.0
            hit_miss_ratio = self.hits / (self.hits + self.misses) if (self.hits + self.misses) > 0 else 0.0
            return {
                'hits': self.hits,
                'misses': self.misses,
                'hit_miss_ratio': f"{hit_miss_ratio:.4f}",
                'max_time': f"{self.max_time_ns / 1e9:.4f} s",
                'min_time': f"{(self.min_time_ns / 1e9) if self.min_time_ns != float('inf') else 0.0:.4f} s",
                'avg_time': f"{avg_time_ns / 1e9:.4f} s",
                'total_calls': self.total_calls,
                'current_size': len(self.cache),
                'max_size': self.cache.maxsize,
                'ttl_sec': f"{self.ttl:.4f}" if self.ttl is not None else "None"
            }

    async def clear(self):
        """Clear the cache and reset all stats."""
        async with self.lock:
            self.cache.clear()
            self.expiry_times.clear()
            self.hits = 0
            self.misses = 0
            self.total_time_ns = 0
            self.total_calls = 0
            self.max_time_ns = 0
            self.min_time_ns = float('inf')




_MAX_SIZE_SENTINEL = object()
_TTL_SENTINEL = object()

def async_cache(maxsize=_MAX_SIZE_SENTINEL, ttl=_TTL_SENTINEL, policy="LFU"):
    """
    A decorator to apply asynchronous caching to a function.

    Args:
        maxsize (int): Maximum size of the cache.
        ttl (Optional[int]): Time-to-Live for cache entries in seconds. Defaults to None.
        policy (str): Cache eviction policy. Can be either "LFU" or "LRU". Defaults to "LFU".
    """
    kwargs = {}
    if maxsize is not _MAX_SIZE_SENTINEL:
        kwargs['maxsize'] = maxsize
    if ttl is not _TTL_SENTINEL:
        kwargs['ttl'] = ttl

    cache_instance = AsyncCache(**kwargs, policy=policy)

    def decorator(fn):
        @functools.wraps(fn)
        async def wrapped_instance(*args, **kwargs):
            if not args:
                # No positional arguments; cache key based only on keyword arguments
                cache_key = make_hashable((None, frozenset(kwargs.items())))
            else:
                # Check if the first argument is a class instance (indicating a bound method)
                b = inspect.isclass(args[0].__class__)
                if b:
                    # The function is a method bound to an instance
                    instance_id = id(args[0])
                    cache_key = make_hashable((instance_id, args[1:], frozenset(kwargs.items())))
                else:
                    # The function is a standalone function
                    cache_key = make_hashable((args, frozenset(kwargs.items())))

            # Try to get the result from the cache
            try:
                value = await cache_instance.__getitem__(cache_key)
                return value
            except KeyError:
                pass
            # Calculate the result and store it in the cache
            start_time_ns = time.perf_counter_ns()
            result = await fn(*args, **kwargs)
            exec_time_ns = time.perf_counter_ns() - start_time_ns
            await cache_instance.update_profiling(exec_time_ns)
            await cache_instance.__setitem__(cache_key, result)
            return result

        wrapped_instance.cache_instance = cache_instance  # Attach the cache instance to the function
        return wrapped_instance

    return decorator



"""
This module is usable to calculate cosine similarity for small number of vectors.

See here https://alex-ber.medium.com/in-memory-similarity-search-998582fbb802 for documentation.

You need to install some 3rd-party dependencies. In order to use it, you should have them installed first.
To do it, run `python -m pip install alex-ber-utils[np]`.

For production purposes, you should install another dependency such as langchain_openai to use, for example, OpenAIEmbeddings.
Run `python -m pip install langchain-openai`.

This module contains SimpleEmbeddings for some simple-minded in-memory calculation of embeddings.
It is provided mainly for educational purposes and for tests. It is not intended to be used in production.

"""

import logging
from typing import List, Tuple, Dict, Protocol

logger = logging.getLogger(__name__)


# For example
# from langchain_openai import OpenAIEmbeddings

class Embeddings(Protocol):
    """
    Protocol for embedding classes.
    """

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """
        Embed a list of documents (texts) into a list of vectors.

        Args:
            texts (List[str]): List of documents to embed.

        Returns:
            List[List[float]]: List of embedding vectors.
        """
        ...


class SimpleEmbeddings:
    """
    Simple in-memory implementation of embeddings for educational and testing purposes.
    """

    def __init__(self, dims: int = 1536):
        """
        Initialize the SimpleEmbeddings with a specified dimension size.

        Args:
            dims (int): Dimension size of the embedding vectors. Default is 1536.
        """
        self.dims = dims

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """
        Embed a list of documents (texts) into a list of vectors.

        Args:
            texts (List[str]): List of documents to embed.

        Returns:
            List[List[float]]: List of embedding vectors.
        """
        embeddings = []
        for text in texts:
            embedding_vector = [0] * self.dims
            # This effectively counts the occurrences of each character in the text, mapped to a fixed-size vector.
            for char in text:
                index = hash(char) % self.dims
                embedding_vector[index] += 1
            embeddings.append(embedding_vector)
        return embeddings


try:
    import numpy as np
except ImportError:
    import warnings

    warning = (
        "You appear to be missing some optional dependencies; "
        "please run 'python -m pip install alex-ber-utils[numpy]'."
    )
    warnings.warn(warning, ImportWarning)
    raise


def _calc_embedding_as_matrix(embeddings: Embeddings, text: str) -> np.ndarray:
    """
    Calculate the embedding of a single text as a matrix.

    Args:
        embeddings (Embeddings): Embedding class instance.
        text (str): Text to embed.

    Returns:
        np.ndarray: Embedding vector as a matrix.
    """
    v = embeddings.embed_documents([text])[0]
    return np.array(v).reshape(1, -1)


def find_most_similar_with_scores(
        embeddings: Embeddings,
        input_text: str,
        /,
        *args: List[str],
        verbose=True
) -> List[Tuple[Tuple[int, str], float]]:
    """
    Find the most similar texts to the input text with their similarity scores.

    Args:
        embeddings (Embeddings): Embedding class instance.
        input_text (str): Input text to compare.
        *args (List[str]): List of texts to compare against.
        verbose (bool): If True, logs additional information. Default is True.

    Returns:
        List[Tuple[Tuple[int, str], float]]: A list of tuples where each tuple contains:
            - A sub-tuple (index, text): index is the position of the text in the input list,
              and text is the corresponding text.
            - A float representing the similarity score.

        If no comparison texts are provided (*args is empty),
        the function returns [((some negative index, input_text), 0.0)].
    """
    logger.info("find_most_similar_with_scores()")
    if not args:
        # List[Tuple[Tuple[int, str], float]]
        # (i, example), score (-1, input_text), 0.0
        return [((-1, input_text), 0.0)]

    input_v: np.ndarray = _calc_embedding_as_matrix(embeddings, input_text)
    example_embeddings_d: Dict[Tuple[int, str], np.ndarray] = {
        (i, example): _calc_embedding_as_matrix(embeddings, example)
        for i, example in enumerate(args)
    }

    # Stack all example embeddings into a single matrix
    example_matrix = np.vstack([v for v in example_embeddings_d.values()])

    # Calculate norms based on the specified ord
    input_norm = np.linalg.norm(input_v, axis=1, keepdims=True)
    example_norms = np.linalg.norm(example_matrix, axis=1, keepdims=True)

    # Calculate cosine similarities in one go
    similarities_matrix = np.dot(input_v, example_matrix.T) / np.outer(input_norm, example_norms)

    # Handle numerical issues
    similarities_matrix[np.isnan(similarities_matrix) | np.isinf(similarities_matrix)] = 0.0

    # Extract scores and sort
    similarities_d: Dict[Tuple[int, str], float] = {key: similarities_matrix[0, idx] for idx, key in enumerate(example_embeddings_d.keys())}
    sorted_similarities_l: List[Tuple[Tuple[int, str], float]] = sorted(similarities_d.items(), key=lambda item: item[1], reverse=True)

    if verbose:
        logger.info(f'Target is {input_text}')
        for (i, example), score in sorted_similarities_l:
            logger.info(f"{i} {example}: has cosine similarity {score:.4f}")

    return sorted_similarities_l


def find_most_similar(
        embeddings: Embeddings,
        input_text: str,
        /,
        *args: List[str],
        verbose=True
) -> Tuple[int, str]:
    """
    Find the most similar text to the input text.

    Args:
        embeddings (Embeddings): Embedding class instance.
        input_text (str): Input text to compare.
        *args (List[str]): List of texts to compare against.
        verbose (bool): If True, logs additional information. Default is True.

    Returns:
        Tuple[int, str]: A tuple containing the index and the most similar text.

    If no comparison texts are provided (*args is empty), the function returns (some negative index, input_text).
    """
    logger.info("find_most_similar()")
    sorted_similarities_l: List[Tuple[Tuple[int, str], float]] = \
        find_most_similar_with_scores(embeddings, input_text,*args, verbose=verbose)
    return sorted_similarities_l[0][0]


import logging
import time
import threading
import pytest
import copy

logger = logging.getLogger(__name__)

from importlib.resources import path

import collections
from collections import OrderedDict
import io
import yaml
from yaml import FullLoader
from alexber.utils.inspects import has_method
from alexber.utils._ymlparsers_extra import format_template


try:
    import alexber.utils.ymlparsers as ymlparsers
    from alexber.utils.ymlparsers import HiYaPyCo
    from alexber.utils.parsers import is_empty
    from jinja2 import DebugUndefined, StrictUndefined, Environment
    from hiyapyco import METHOD_SUBSTITUTE, METHOD_SIMPLE
except ImportError:
    pass

def _reset_ymlparsers():
    ymlparsers.HiYaPyCo.jinja2Lock = None
    ymlparsers.HiYaPyCo.jinja2ctx = None
    ymlparsers._load_d = None
    ymlparsers._safe_dump_d = None

@pytest.fixture
@pytest.mark.yml
def ymlparsersSetup(mocker):
    _reset_ymlparsers()

    ymlparsers.initConfig()

@pytest.fixture(scope='session')
@pytest.mark.yml
def exp_config_d():
    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

    with path(pck, 'config.yml') as full_path:
        with open(full_path, 'r') as f:
            exp_d = yaml.load(f, FullLoader)
    return exp_d



@pytest.fixture
@pytest.mark.yml
def ymlparsersCleanup(mocker):
    yield None
    _reset_ymlparsers()

class DymmyLock(object):
    def acquire(self, blocking=True):
        pass

    def release(self):
        pass

def create_mock_lock(mocker):
    mock_lock = DymmyLock()

    def __enter__():
        mock_lock.acquire()
        return mock_lock

    def __exit__(t, v, tb):
        mock_lock.release()

    mocker.patch.object(mock_lock, 'acquire', autospec=True, spec_set=True)
    mocker.patch.object(mock_lock, 'release', autospec=True, spec_set=True)
    mocker.patch.object(type(mock_lock), '__enter__', side_effect=__enter__, create=True)
    mocker.patch.object(type(mock_lock), '__exit__', side_effect=__exit__, create=True)
    return mock_lock


class TestYmlparsersInit(object):
    @pytest.mark.yml
    def test_initConfig_default_params(self, request, mocker, ymlparsersCleanup):
        logger.info(f'{request._pyfuncitem.name}()')
        ymlparsers.initConfig()
        lock = HiYaPyCo.jinja2Lock
        pytest.assume(lock is not None)
        lock_cls = type(lock)
        b = has_method(lock_cls, 'acquire')
        pytest.assume(b)
        b = has_method(lock_cls, 'release')
        pytest.assume(b)

        jinja2Env = HiYaPyCo.jinja2ctx
        pytest.assume(jinja2Env is not None)

        jinja2EnvUndefined = jinja2Env.undefined
        b = issubclass(jinja2EnvUndefined,DebugUndefined)
        pytest.assume(b)

        jinja2EnvGloblas = jinja2Env.globals
        b = 'uname' in jinja2EnvGloblas
        pytest.assume(b)

        _load_d = ymlparsers._load_d
        assert _load_d is not None
        b = isinstance(_load_d, collections.abc.Mapping)
        pytest.assume(b)

        for k, exp_v in {'method':METHOD_SUBSTITUTE,
                     'mergelists':False,
                     'interpolate':True,
                     'castinterpolated':True}.items():
            v = _load_d[k]
            pytest.assume(exp_v==v)


        _safe_dump_d = ymlparsers._safe_dump_d
        assert _safe_dump_d is not None
        b= isinstance(_safe_dump_d, collections.abc.Mapping)
        pytest.assume(b)

        for k, exp_v in {'default_flow_style':False,
                         'sort_keys':False}.items():
            v = _safe_dump_d[k]
            pytest.assume(exp_v==v)

        pytest.assume("TODO: HiYaPyCo._substmerge() bug workarround, see https://github.com/zerwes/hiyapyco/pull/38",
                      HiYaPyCo._deepmerge==HiYaPyCo._substmerge)

    @pytest.mark.yml
    def test_initConfig_other_params(self, request, mocker, ymlparsersCleanup):
        logger.info(f'{request._pyfuncitem.name}()')

        mock_lock = create_mock_lock(mocker)

        p_load = {'method':METHOD_SIMPLE,
               'mergelists':True,
               'interpolate':False,
               'castinterpolated':False}

        p_safe_dump = {'default_flow_style':True,
                       'sort_keys':True}

        p_undefined = StrictUndefined

        dumb = lambda:None

        p_globals = {"foo":dumb,
                     "uname":dumb}

        ymlparsers.initConfig(jinja2Lock=mock_lock,
                              jinja2ctx={'undefined':p_undefined,
                                         'globals':p_globals
                                         },
                              load=p_load,
                              safe_dump=p_safe_dump
                              )
        lock = HiYaPyCo.jinja2Lock
        pytest.assume(lock is not None)
        pytest.assume(lock==mock_lock)

        pytest.assume(mock_lock.acquire.call_count > 0)
        pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

        jinja2Env = HiYaPyCo.jinja2ctx
        pytest.assume(jinja2Env is not None)

        jinja2EnvUndefined = jinja2Env.undefined
        b = issubclass(jinja2EnvUndefined, p_undefined)
        pytest.assume(b)

        jinja2EnvGloblas = jinja2Env.globals
        f= jinja2EnvGloblas['uname']
        pytest.assume(f==dumb)
        f = jinja2EnvGloblas['foo']
        pytest.assume(f == dumb)

        _load_d = ymlparsers._load_d
        assert _load_d is not None
        b = isinstance(_load_d, collections.abc.Mapping)
        pytest.assume(b)

        for k, exp_v in p_load.items():
            v = _load_d[k]
            pytest.assume(exp_v ==v)

        _safe_dump_d = ymlparsers._safe_dump_d
        assert _safe_dump_d is not None
        b = isinstance(_safe_dump_d, collections.abc.Mapping)
        pytest.assume(b)

        for k, exp_v in p_safe_dump.items():
            v = _safe_dump_d[k]
            pytest.assume(exp_v == v)

class TestDisableVarSubst(object):

    @pytest.mark.yml
    def test_intented_usage(self, request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
        logger.info(f'{request._pyfuncitem.name}()')

        mock_lock = create_mock_lock(mocker)

        orig_jinja2ctx = HiYaPyCo.jinja2ctx

        mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)
        jinja2ctx_mock = mocker.patch.object(HiYaPyCo, 'jinja2ctx', spec_set=True)

        mock_variable_start_string = mocker.PropertyMock(return_value=orig_jinja2ctx.variable_start_string)
        type(jinja2ctx_mock).variable_start_string = mock_variable_start_string
        mock_variable_end_string = mocker.PropertyMock(return_value=orig_jinja2ctx.variable_end_string)
        type(jinja2ctx_mock).variable_end_string = mock_variable_end_string
        mock_block_start_string = mocker.PropertyMock(return_value=orig_jinja2ctx.block_start_string)
        type(jinja2ctx_mock).block_start_string = mock_block_start_string
        mock_block_end_string = mocker.PropertyMock(return_value=orig_jinja2ctx.block_end_string)
        type(jinja2ctx_mock).block_end_string = mock_block_end_string

        mocks = [mock_variable_start_string, mock_variable_end_string, mock_block_start_string, mock_block_end_string]

        # orig_exit = ymlparsers.DisableVarSubst.__exit__
        # mock_exit = mocker.patch.object(ymlparsers.DisableVarSubst, '__exit__', side_effect=orig_exit,
        #                                 autospec=True, spec_set=True)



        with ymlparsers.DisableVarSubst():
            logger.debug("dff")
            pass

        # pytest.assume(mock_exit.call_count == 1)
        pytest.assume(mock_lock.acquire.call_count > 0)
        pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

        setter_called = None
        for mock in mocks:
            setter_called = None
            pytest.assume(mock.call_count > 0)
            for kall in mock.call_args_list:
                #(param,), _ = kall
                args, _ = kall
                if not is_empty(args) and '|' in args[0]:
                    setter_called = True
                    break
            pytest.assume(setter_called)

        pytest.assume(jinja2ctx_mock.variable_start_string == orig_jinja2ctx.variable_start_string)
        pytest.assume(jinja2ctx_mock.variable_end_string == orig_jinja2ctx.variable_end_string)
        pytest.assume(jinja2ctx_mock.block_start_string == orig_jinja2ctx.block_start_string)
        pytest.assume(jinja2ctx_mock.block_end_string == orig_jinja2ctx.block_end_string)

    class DummyEnvironment(object):
        def __init__(self, *args, **kwargs):
            self.delegate = kwargs['delegate']
            self._variable_start_string = self.delegate.variable_start_string
            self._variable_end_string = self.delegate.variable_end_string
            self._block_start_string =  self.delegate.block_start_string
            self._block_end_string = self.delegate.block_end_string
            self.raiseAlways = kwargs.get('raiseAlways', False)

        @property
        def variable_start_string(self):
            return self._variable_start_string

        @variable_start_string.setter
        def variable_start_string(self, new_name):
            self._variable_start_string = new_name

        @property
        def variable_end_string(self):
            return self._variable_end_string

        @variable_end_string.setter
        def variable_end_string(self, new_name):
            self._variable_end_string = new_name

        @property
        def block_start_string(self):
            return self._block_start_string

        @block_start_string.setter
        def block_start_string(self, new_name):
            self._block_start_string = new_name

        @property
        def block_end_string(self):
            return self._block_end_string

        @block_end_string.setter
        def block_end_string(self, new_name):
            if self.raiseAlways or ('|' not in new_name):
                raise ValueError
            self._block_end_string = new_name

    @pytest.mark.yml
    def test_exception_in_exit(self, request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
        logger.info(f'{request._pyfuncitem.name}()')
        #I check 2 things
        #1. Explicately passing params
        #2. Releasing lock even if exception occures in DisableVarSubst. __exit__

        mock_lock = create_mock_lock(mocker)

        orig_jinja2ctx = HiYaPyCo.jinja2ctx
        duumy_jinja2ctx = TestDisableVarSubst.DummyEnvironment(delegate=orig_jinja2ctx)

        #orig_exit = ymlparsers.DisableVarSubst.__exit__
        # mock_exit = mocker.patch.object(ymlparsers.DisableVarSubst, '__exit__', side_effect=orig_exit,
        #                                 autospec=True, spec_set=True)

        with pytest.raises(ValueError):
            with ymlparsers.DisableVarSubst(jinja2ctx=duumy_jinja2ctx, jinja2Lock=mock_lock):
                pass

        # pytest.assume(mock_exit.call_count == 1)
        pytest.assume(mock_lock.acquire.call_count > 0)
        pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

        pytest.assume(duumy_jinja2ctx != orig_jinja2ctx)
        pytest.assume(mock_lock != orig_jinja2ctx)

        pytest.assume(duumy_jinja2ctx.variable_start_string == orig_jinja2ctx.variable_start_string)
        pytest.assume(duumy_jinja2ctx.variable_end_string == orig_jinja2ctx.variable_end_string)
        pytest.assume(duumy_jinja2ctx.block_start_string == orig_jinja2ctx.block_start_string)
        pytest.assume(duumy_jinja2ctx.block_end_string !=  orig_jinja2ctx.block_end_string )

    @pytest.mark.yml
    def test_exception_in_enter(self, request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
        logger.info(f'{request._pyfuncitem.name}()')
        #I check 2 things
        #1. Explicately passing params
        #2. Releasing lock even if exception occures in DisableVarSubst.__enter__

        mock_lock = create_mock_lock(mocker)

        orig_jinja2ctx = HiYaPyCo.jinja2ctx
        duumy_jinja2ctx = TestDisableVarSubst.DummyEnvironment(delegate=orig_jinja2ctx, raiseAlways=True)

        # orig_exit = ymlparsers.DisableVarSubst.__exit__
        # mock_exit = mocker.patch.object(ymlparsers.DisableVarSubst, '__exit__', side_effect=orig_exit,
        #                     autospec=True, spec_set=True)


        with pytest.raises(ValueError):
            with ymlparsers.DisableVarSubst(jinja2ctx=duumy_jinja2ctx, jinja2Lock=mock_lock):
                pass

        # pytest.assume(mock_exit.call_count == 0)
        pytest.assume(mock_lock.acquire.call_count > 0)
        pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

        pytest.assume(duumy_jinja2ctx != orig_jinja2ctx)
        pytest.assume(mock_lock != orig_jinja2ctx)

        pytest.assume(duumy_jinja2ctx.variable_start_string != orig_jinja2ctx.variable_start_string)
        pytest.assume(duumy_jinja2ctx.variable_end_string != orig_jinja2ctx.variable_end_string)
        pytest.assume(duumy_jinja2ctx.block_start_string != orig_jinja2ctx.block_start_string)
        pytest.assume(duumy_jinja2ctx.block_end_string == orig_jinja2ctx.block_end_string)

    @pytest.mark.yml
    def test_exception_in_code(self, request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
        logger.info(f'{request._pyfuncitem.name}()')
        #I check 2 things
        #1. Explicately passing params
        #2. Releasing lock even if exception occures in the with

        mock_lock = create_mock_lock(mocker)

        orig_jinja2ctx = HiYaPyCo.jinja2ctx

        mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)
        jinja2ctx_mock = mocker.patch.object(HiYaPyCo, 'jinja2ctx', spec_set=True)

        mock_variable_start_string = mocker.PropertyMock(return_value=orig_jinja2ctx.variable_start_string)
        type(jinja2ctx_mock).variable_start_string = mock_variable_start_string
        mock_variable_end_string = mocker.PropertyMock(return_value=orig_jinja2ctx.variable_end_string)
        type(jinja2ctx_mock).variable_end_string = mock_variable_end_string
        mock_block_start_string = mocker.PropertyMock(return_value=orig_jinja2ctx.block_start_string)
        type(jinja2ctx_mock).block_start_string = mock_block_start_string
        mock_block_end_string = mocker.PropertyMock(return_value=orig_jinja2ctx.block_end_string)
        type(jinja2ctx_mock).block_end_string = mock_block_end_string

        mocks = [mock_variable_start_string, mock_variable_end_string, mock_block_start_string, mock_block_end_string]
        # orig_exit = ymlparsers.DisableVarSubst.__exit__
        # mock_exit = mocker.patch.object(ymlparsers.DisableVarSubst, '__exit__', side_effect=orig_exit,
        #                                 autospec=True, spec_set=True)

        with pytest.raises(ValueError):
            with ymlparsers.DisableVarSubst():
                raise ValueError

        # pytest.assume(mock_exit.call_count == 1)
        pytest.assume(mock_lock.acquire.call_count > 0)
        pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

        setter_called = None
        for mock in mocks:
            setter_called = None
            pytest.assume(mock.call_count > 0)
            for kall in mock.call_args_list:
                # (param,), _ = kall
                args, _ = kall
                if not is_empty(args) and '|' in args[0]:
                    setter_called = True
                    break
            pytest.assume(setter_called)

        pytest.assume(jinja2ctx_mock.variable_start_string == orig_jinja2ctx.variable_start_string)
        pytest.assume(jinja2ctx_mock.variable_end_string == orig_jinja2ctx.variable_end_string)
        pytest.assume(jinja2ctx_mock.block_start_string == orig_jinja2ctx.block_start_string)
        pytest.assume(jinja2ctx_mock.block_end_string == orig_jinja2ctx.block_end_string)


@pytest.mark.yml
def test_ymlparsers_load_single_no_substition(request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')

    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)

    with path(pck, 'config.yml') as full_path:
        with ymlparsers.DisableVarSubst():
            default_d = ymlparsers.load([str(full_path)])

    # ymlparsers.load() when uses HiYaPyCo.jinja2ctx uses HiYaPyCo.jinja2Lock
    # note, that ymlparsers.DisableVarSubst also use 1 time HiYaPyCo.jinja2Lock
    pytest.assume(mock_lock.acquire.call_count > 1)
    pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)


    app_d = default_d.get('app', None)
    exp_d = copy.deepcopy(exp_config_d)
    exp_app_d = exp_d.get('app', None)

    inner_host_name = app_d.get('inner_host_name', None)
    exp_host_name = exp_app_d.get('inner_host_name', None)
    pytest.assume(exp_host_name==inner_host_name)
    cli_template = app_d.get('cli_template')
    pytest.assume('inner_host_name' in cli_template)
    pytest.assume(exp_config_d==default_d)


@pytest.mark.yml
def test_ymlparsers_load_single_with_substition(request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')

    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])


    mock_lock = create_mock_lock(mocker)
    mocker.patch.object(HiYaPyCo, 'jinja2Lock', new=mock_lock, spec_set=True)

    with path(pck, 'config.yml') as full_path:
        default_d = ymlparsers.load([str(full_path)])

    pytest.assume(mock_lock.acquire.call_count > 1)
    pytest.assume(mock_lock.release.call_count == mock_lock.acquire.call_count)

    app_d = default_d['app']
    exp_d = copy.deepcopy(exp_config_d)
    exp_app_d = exp_d.get('app', None)

    inner_host_name = app_d.get('inner_host_name', None)
    exp_host_name = exp_app_d.get('inner_host_name', None)
    pytest.assume(exp_host_name==inner_host_name)

    exp_cli_template = exp_app_d.get('cli_template', None)
    exp_cli_template = format_template(exp_cli_template, app_inner_host_name=exp_host_name)
    exp_app_d['cli_template']=exp_cli_template


    cli_template = app_d.get('cli_template')
    pytest.assume('inner_host_name' not in cli_template)

    cli_template = app_d['cli_template']
    pytest.assume('inner_host_name' not in cli_template)
    pytest.assume(exp_d==default_d)

    white_list = app_d.get('white_list', None)
    exp_white_list = exp_app_d.get('white_list', None)
    pytest.assume(exp_white_list==white_list)


#order is non-alephabetic intentionally
_data_list = {'profiles': ['local', 'dev']}
_data_dict = {'general':
                 { 'whiteListSysOverrideKeys': ['app'],
                   'profiles': ['local', 'dev']},
              'app':
                  {'host_name': 'google.com'},
             }

@pytest.mark.parametrize(
     'data, kwds',

    [
        (_data_list, {}),

        #sort_key is ignored
        (_data_list, {'default_flow_style': None,  'sort_keys': True}),
        (_data_list, {'default_flow_style': False, 'sort_keys': True}),
        (_data_list, {'default_flow_style': True,  'sort_keys': True}),

        (_data_dict, {'default_flow_style': None,  'sort_keys': True}),
        (_data_dict, {'default_flow_style': False, 'sort_keys': True}),
        (_data_dict, {'default_flow_style': True,  'sort_keys': True}),

    ]
)
@pytest.mark.yml
def test_safe_dump_composite(request, mocker, ymlparsersSetup, ymlparsersCleanup, data, kwds):
    logger.info(f'{request._pyfuncitem.name}()')

    default_flow_style = kwds.get('default_flow_style', False)

    with io.StringIO() as buf:
        ymlparsers.safe_dump(data, stream=buf, **kwds)
        value = buf.getvalue()
        logger.debug(value)
        if default_flow_style is not None and not default_flow_style:
            pytest.assume('-' in value)
        else:
            pytest.assume('[' in value)
            pytest.assume(']' in value)

        with ymlparsers.DisableVarSubst():
            restored_d = ymlparsers.load(value)

    pytest.assume(data == restored_d)

@pytest.mark.yml
def test_safe_dump(request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')

    orig_d = dict(exp_config_d)
    d = OrderedDict(exp_config_d)
    d['general'] = OrderedDict(d['general'])
    d['app'] = OrderedDict(d['app'])

    with io.StringIO() as buf:
        ymlparsers.safe_dump(d, stream=buf)
        value = buf.getvalue()

    with ymlparsers.DisableVarSubst():
        restored_d = ymlparsers.load(value)

    pytest.assume(orig_d==restored_d)

@pytest.mark.yml
def test_as_str(request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')

    orig_d = dict(exp_config_d)
    d = OrderedDict(exp_config_d)
    d['general'] = OrderedDict(d['general'])
    d['app'] = OrderedDict(d['app'])

    orig_safe_dump = ymlparsers.safe_dump
    mock_safe_dump = mocker.patch.object(ymlparsers, 'safe_dump', side_effect=orig_safe_dump,
                                    autospec=True, spec_set=True)

    value = ymlparsers.as_str(d)
    pytest.assume(mock_safe_dump.call_count > 0)

    with ymlparsers.DisableVarSubst():
        restored_d = ymlparsers.load(value)

    pytest.assume(orig_d==restored_d)

@pytest.mark.yml
def test_ymlparsers_load_multiple_no_substition(request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')

    exp_d = copy.deepcopy(exp_config_d)
    exp_d['general']['log']['formatters']['detail'] = {'format': \
                                                '%(asctime)-14s %(levelname)s [%(name)s.%(funcName)s] %(message)s',
                                                       'datefmt': \
                                                '%Y-%m-%d %H:%M:%S'}
    exp_d['general']['log']['root']['level'] = 'INFO'
    exp_d['app']['inner_host_name'] = 'yahoo.com'
    exp_d['app']['white_list'] = ['gamma', 'alpha', 'betha']
    exp_d['app']['alt_white_list'] = ['c', 'b', 'a']



    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

    with path(pck, 'config.yml') as full_path:
        with path(pck, 'config-dev.yml') as full_dev_path:
            with ymlparsers.DisableVarSubst():
                default_d = ymlparsers.load([str(full_path), str(full_dev_path)])


    app_d = default_d.get('app', None)
    exp_app_d = exp_d.get('app', None)

    inner_host_name = app_d.get('inner_host_name', None)
    exp_host_name = exp_app_d.get('inner_host_name', None)
    pytest.assume(exp_host_name==inner_host_name)

    cli_template = app_d.get('cli_template')
    pytest.assume('inner_host_name' in cli_template)

    pytest.assume(exp_d==default_d)

@pytest.mark.yml
def test_ymlparsers_load_multiple_with_substition(request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')

    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

    exp_d = copy.deepcopy(exp_config_d)
    exp_d['general']['log']['formatters']['detail'] = {'format': \
                                                           '%(asctime)-14s %(levelname)s [%(name)s.%(funcName)s] %(message)s',
                                                       'datefmt': \
                                                           '%Y-%m-%d %H:%M:%S'}
    exp_d['general']['log']['root']['level'] = 'INFO'
    exp_d['app']['inner_host_name'] = 'yahoo.com'
    exp_d['app']['white_list'] = ['gamma', 'alpha', 'betha']
    exp_d['app']['alt_white_list'] = ['c', 'b', 'a']

    with path(pck, 'config.yml') as full_path:
        with path(pck, 'config-dev.yml') as full_dev_path:
            default_d = ymlparsers.load([str(full_path), str(full_dev_path)])


    app_d = default_d['app']
    exp_app_d = exp_d.get('app', None)

    inner_host_name = app_d.get('inner_host_name', None)
    exp_host_name = exp_app_d.get('inner_host_name', None)
    pytest.assume(exp_host_name==inner_host_name)

    exp_cli_template = exp_app_d.get('cli_template', None)
    exp_cli_template = format_template(exp_cli_template, app_inner_host_name=exp_host_name)
    exp_app_d['cli_template']=exp_cli_template


    cli_template = app_d.get('cli_template')
    pytest.assume('inner_host_name' not in cli_template)

    cli_template = app_d['cli_template']
    pytest.assume('inner_host_name' not in cli_template)

    pytest.assume(exp_d==default_d)


def _run_without_substition(content, exp_config_d, stop):
    for i in range(stop):
        with ymlparsers.DisableVarSubst():
            default_d = ymlparsers.load([str(content)])

            app_d = default_d.get('app', None)
            exp_app_d = exp_config_d.get('app', None)

            inner_host_name = app_d.get('inner_host_name', None)
            exp_host_name = exp_app_d.get('inner_host_name', None)
            pytest.assume(exp_host_name==inner_host_name)
            cli_template = app_d.get('cli_template')
            pytest.assume('inner_host_name' in cli_template)
            pytest.assume(exp_config_d == default_d)

def _run_with_substition(content, exp_config_d, stop):
    for i in range(stop):
        default_d = ymlparsers.load([str(content)])

        app_d = default_d['app']
        exp_app_d = exp_config_d.get('app', None)

        inner_host_name = app_d.get('inner_host_name', None)
        exp_host_name = exp_app_d.get('inner_host_name', None)
        pytest.assume(exp_host_name==inner_host_name)

        exp_cli_template = exp_app_d.get('cli_template', None)
        exp_cli_template = format_template(exp_cli_template, app_inner_host_name=exp_host_name)

        exp_app_d['cli_template'] = exp_cli_template

        cli_template = app_d.get('cli_template')
        pytest.assume('inner_host_name' not in cli_template)

        cli_template = app_d['cli_template']
        pytest.assume('inner_host_name' not in cli_template)
        pytest.assume(exp_config_d == default_d)

@pytest.mark.yml
def test_ymlparsers_load_it(request, mocker, ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')
    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

    with path(pck, 'config.yml') as full_path:
        with open(full_path, 'r') as f:
            content = f.read()

    stop = 10

    exp_d = copy.deepcopy(exp_config_d)
    exp_app_d = dict(exp_d.get('app', None))
    if exp_app_d is not None:
        exp_d['app'] = exp_app_d

    exp_host_name = exp_app_d.get('inner_host_name', None)
    exp_cli_template = exp_app_d.get('cli_template', None)
    exp_cli_template = format_template(exp_cli_template, app_inner_host_name=exp_host_name)
    exp_app_d['cli_template'] = exp_cli_template


    th1 = threading.Thread(name="run_with_substition",
        target=_run_with_substition, args=(content, exp_d, stop))

    th2 = threading.Thread(name="run_without_substition",
                           target=_run_without_substition, args=(str(content), copy.deepcopy(exp_config_d), stop))

    th1.start()
    time.sleep(2)
    th2.start()
    th1.join()
    th2.join()


if __name__ == "__main__":
    pytest.main([__file__])



import logging
import sys
from . importer import importer

# This modules is based upon https://github.com/fx-kirin/py-stdlogging/blob/master/stdlogging.py
# See also https://github.com/fx-kirin/py-stdlogging/pull/1.
# I have encountered on this module here
# https://stackoverflow.com/questions/47325506/making-python-loggers-log-all-stdout-and-stderr-messages 
# Quote: "But be careful to capture stdout because it's very fragile"

def _validate_param(d, param_name):
    if d is None:
        raise ValueError(f"Expected '{param_name} param not found")

class StreamToLogger:
    """
    This is adapter class from any stream-like object to logging.Logger.
    """

    def __init__(self, **kwargs):
        """
        It is recommended to use package-level initStream() function and not this method directly.
        :param logger: Required. Standard Python's Logger or any Logger-like object.
        :param stream: Required. sys.stderr, sys.stdout or any other stream-like object.
        :param log_level: Optional. If not supplied, logging.DEBUG will be used.
        """

        _validate_param(kwargs, 'logger')
        _validate_param(kwargs, 'log_level')
        _validate_param(kwargs, 'stream')

        self.logger = kwargs.pop('logger')
        self.log_level = kwargs.pop('log_level', logging.DEBUG)
        self.stream = kwargs.pop('stream')

    def write(self, lines):
        if lines:
            lines = lines+'\n'
            for line in lines.split('\n'):
                if line:
                    self.logger.log(self.log_level, line.rstrip())

    def flush(self):
        self.stream.flush()



def initStream(logger=None, logger_level=logging.ERROR, stream_getter=None, stream_setter=None, adapter_cls=None):
    """
    Preffered API.
    stream_getter() is supplier/factory method that returns stream-like object (i.e. sys.stderr) that we're adapting upon.
                   It's intended usage is to supply stream-like object that we want to apapt upon.
    stream_setter() is consumer method that receives wrapped object as parameter.
                   It's intended usage is to overwrite source stream-like, i.e. sys.stderr = s.
    :param logger: Optional. If not supplied logging.getLogger('stderr') will be used.
    :param logger_level: Optional. If not supplied logging.ERROR will be used.
    :param stream_getter: Optional. if not supplied method that returns sys.stderr will be used.
    :param stream_setter: Optional. if not supplied method that get's strema-like object and set's sys.stderr will be used.
    :param adapter_cls: Optional. Can be str or class. if not supplied than StreamToLogger is used.
    :return:
    """
    if logger is None:
        logger = logging.getLogger('stderr')

    if stream_getter is None:
        def stream_getter():
            return sys.stderr

    if stream_setter is None:
        def stream_setter(s):
            sys.stderr = s

    stream = stream_getter()

    if adapter_cls is None:
        adapter_cls = StreamToLogger
    elif isinstance(adapter_cls, str):
        adapter_cls = importer(adapter_cls)

    sl = adapter_cls(logger=logger, stream=stream, logger_level=logger_level)
    stream_setter(sl)


import os

import setuptools
from setuptools import setup


#VERSION should be defined before importing UploadCommand
VERSION = '0.12.10b'
from alexber.utils import UploadCommand
NAME = 'alex_ber_utils'
SHORT_NAME = 'utils'
VCS_URL = 'https://github.com/alex-ber/AlexBerUtils'
DESCRIPTION = 'AlexBerUtils is collection of the small utilities'
AUTHOR = 'Alexander Berkovich'


base_dir = os.path.dirname(os.path.realpath(__file__))

def get_content(filename):
    with open(os.path.join(base_dir, filename)) as f:
        content = f.read().splitlines()
    return content

install_requires = get_content('requirements.txt')
tests_require = get_content('requirements-tests.txt')

extras = {
    'fabric' : get_content('req-fabric.txt'),
    'yaml' : get_content('req-yml.txt'),
    'yml' : get_content('req-yml.txt'),
    'env' : get_content('req-env.txt'),
    'np' : get_content('req-np.txt'),
    'tests': tests_require,
    'piptools' : get_content('req-piptools.txt')
}

lnk_data = os.path.join('alexber', SHORT_NAME, 'data')



try:
    try:
        os.unlink(lnk_data)
    except OSError:
        pass

    os.symlink(os.path.join('..', '..', 'data'), lnk_data)

    setup(
        name=NAME,
        version=VERSION,
        url=VCS_URL,
        author=AUTHOR,
        description=DESCRIPTION,
        long_description="\n\n".join([
            open(os.path.join(base_dir, "README.md"), "r").read(),
            open(os.path.join(base_dir, "CHANGELOG.md"), "r").read()
        ]),
        long_description_content_type="text/markdown",
        packages=setuptools.find_packages(exclude=('tests', 'tests.*', 'data')),
        # see https://stackoverflow.com/a/26533921
        # see also https://stackoverflow.com/questions/24347450/how-do-you-add-additional-files-to-a-wheel
        # data_files=[(f'Lib/site-packages/alexber/{SHORT_NAME}', ['data/config.yml', 'data/requirements-src.txt',
        #                                                    'data/driver.py']),
        #             #(f'lib/python3.7/site-packages/alexber/{SHORT_NAME}', ['requirements-src.txt'])
        #             ],
        # package_data={'alexber.{SHORT_NAME}': ['data/*', 'data/config.yml',
        #                                   'data/requirements-stc.txt', 'data/requirements-dest.txt']},
        package_data={f'alexber.{SHORT_NAME}': ['data/*'
                                                ]},
        include_package_data=True,
        install_requires=install_requires,
        # entry_points={"console_scripts": [
        #     f"python-{SHORT_NAME}-tool=alexber.{SHORT_NAME}.data.__main__:main"
        # ]},
        # $ setup.py publish support.
        # python3 setup.py upload
        cmdclass={
            'upload': UploadCommand,
        },
        extras_require=extras,
        test_suite="tests",
        tests_require=tests_require,
        setup_requires=['pytest-runner'],
        namespace_packages=('alexber',),
        license='Apache 2.0',
        keywords='tools tool utils enum enums threadlocal UploadCommand upload uuid1mc uuid UUID UUID1 UUID4 UU1DMC ' \
                 'issetdescriptor ismethod ' \
                 'importer new_instance safe_eval is_empty parse_boolean Properties java.util.Properties',
        classifiers=[
            # See: https://pypi.python.org/pypi?:action=list_classifiers
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'License :: OSI Approved :: BSD License',

            # List of python versions and their support status:
            # https://en.wikipedia.org/wiki/CPython#Version_history
            'Programming Language :: Python',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'Programming Language :: Python :: Implementation :: CPython',
            "Topic :: Utilities",
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Desktop Environment',
            'Topic :: Education',
            'Operating System :: OS Independent',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX',
            'Operating System :: Unix',
            'Natural Language :: English',
        ],
        python_requires='>=3.8',
        zip_safe=False,

    )

finally:
    try:
        os.unlink(lnk_data)
    except OSError:
        pass



import logging
import pytest
from alexber.utils.pprint import pformat

logger = logging.getLogger(__name__)

def assume(s1, s2):
    # Remove all whitespace characters from both strings
    s1_no_whitespace = ''.join(s1.split())
    s2_no_whitespace = ''.join(s2.split())


    pytest.assume(s1_no_whitespace == s2_no_whitespace)

input_d = {
        "more_info": [{
            "name": "Arthur",
            "last_name": "Doe",
        },
        {
            "name": "Sherlock",
            "last_name": "Holmes",
        }
        ],
        "answer": "The chance is zero for this",
    }


def test_standard_pprint(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    from pprint import pformat

    actual_s = pformat(input_d)
    exp_value_s = """{'answer': 'The chance is zero for this',
                     'more_info': [{'last_name': 'Doe', 'name': 'Arthur'},
                                   {'last_name': 'Holmes', 'name': 'Sherlock'}]}"""
    assume(exp_value_s, actual_s)

def test_standard_pprint_sort_dicts_false(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    from pprint import pformat

    actual_s = pformat(input_d, sort_dicts=False)
    exp_value_s = """{'more_info': [{'name': 'Arthur', 'last_name': 'Doe'},
                                    {'name': 'Sherlock', 'last_name': 'Holmes'}],
                     'answer': 'The chance is zero for this'}"""
    assume(exp_value_s, actual_s)

def test_my_pprint(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    actual_s = pformat(input_d)
    exp_value_s = """{'more_info': [{'name': 'Arthur', 'last_name': 'Doe'},
                                    {'name': 'Sherlock', 'last_name': 'Holmes'}],
                     'answer': 'The chance is zero for this'}"""
    assume(exp_value_s, actual_s)

def test_my_pprint_sort_dicts_false(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    actual_s = pformat(input_d, sort_dicts=False)
    exp_value_s = """{'more_info': [{'name': 'Arthur', 'last_name': 'Doe'},
                                    {'name': 'Sherlock', 'last_name': 'Holmes'}],
                     'answer': 'The chance is zero for this'}"""
    assume(exp_value_s, actual_s)

def test_my_pprint_sort_dicts_true(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    actual_s = pformat(input_d, sort_dicts=True)
    exp_value_s = """{'answer': 'The chance is zero for this',
                     'more_info': [{'last_name': 'Doe', 'name': 'Arthur'},
                                   {'last_name': 'Holmes', 'name': 'Sherlock'}]}"""
    assume(exp_value_s, actual_s)


if __name__ == "__main__":
    pytest.main([__file__])



import logging
import pytest

try:
    from alexber.utils.in_memory_similarity_search import SimpleEmbeddings, find_most_similar
except ImportError:
    pass

logger = logging.getLogger(__name__)


@pytest.fixture(scope='module')
@pytest.mark.np
def embeddings():
    return SimpleEmbeddings()

@pytest.mark.np
@pytest.mark.parametrize(
    'question, r, a0, a1, expected_idx',

    [
        ("Who was the top goal scorer in the Premier League last season?",
         'According to the given chat history, the top goal scorer in the Premier League last season was Harry Kane, with 23 goals.',

         # 0.9682
         "I don't know who the top goal scorer in the Premier League last season was.",
         # 0.9786
         "The top goal scorer in the Premier League last season was Harry Kane, with 23 goals.",
         1),

        ("Which team won the UEFA Champions League two years ago?",
         'Based on the provided chat history, the team that won the UEFA Champions League two years ago was Bayern Munich.',

         #0.9826
         'The team that won the UEFA Champions League two years ago was Bayern Munich.',
         #0.9592
         "I don't know which team won the UEFA Champions League two years ago.",
         0),


        ("Which country hosted the FIFA World Cup four years ago?",
         'Surveys indicate that 60% of soccer fans believe Lionel Messi is the greatest player of all time.',

         #0.9886
         "60% of soccer fans believe Lionel Messi is the greatest player of all time",
         #0.9802
         "I don't know what percentage of soccer fans believe Lionel Messi is the greatest player of all time.",
         0),

        ("Who is known as the 'King of Football'?",
         'Pelé is known as the "King of Football".',

         #0.9374
         'The "King of Football" is Pele.',
         #0.9045
         'The "King of Football" is Maradona, no, sorry, Pele.',
         0),




        ("Who was named the best goalkeeper in the world last year?",
        'According to sports analysts, the best goalkeeper in the world last year was Manuel Neuer.',

          #0.9571
         'The best goalkeeper in the world last year was Manuel Neuer.',
          #0.9488,
          "I don't know who was named the best goalkeeper in the world last year.",
         0),

        ("What is the most watched soccer match in history?",
         'Records indicate that the most watched soccer match in history is the 2014 FIFA World Cup final between Germany and Argentina',

         #0.9365
         "I don't know what the most watched soccer match in history is.",
         #0.9918
         "The most watched soccer match in history is the 2014 FIFA World Cup final between Germany and Argentina.",
         1),

        ("Which player has the most assists in La Liga history?",
         'Statistics show that the player with the most assists in La Liga history is Lionel Messi.',

         #0.9786
         'The player with the most assists in La Liga history is Lionel Messi.',
         #0.9584
         "I don't know which player has the most assists in La Liga history.",
         0),

        ("What percentage of soccer fans prefer watching club football over international football?",
         'Polls reveal that 70% of soccer fans prefer watching club football over international football.',

         #0.9733
         "I don't know what percentage of soccer fans prefer watching club football over international football.",
         #0.9890
         "70% of soccer fans prefer watching club football over international football",
         1),

        ("Which team has won the most Serie A titles?",
         'Historical data shows that the team that has won the most Serie A titles is Juventus.',

         #0.9764
         'The team that has won the most Serie A titles is Juventus.',
         #0.9380
         "I don't know which team has won the most Serie A titles.",
         0),

        ("Who was awarded the Ballon d'Or last year?",
         "Reports confirm that the player awarded the Ballon d'Or last year was Robert Lewandowski.",

         #0.9403
         "I don't know who was awarded the Ballon d'Or last year.",
         #0.9768
         "The player awarded the Ballon d'Or last year was Robert Lewandowski.",
         1),
    ]
)
def test_find_most_similar(request, embeddings, question, r, a0, a1, expected_idx):
    logger.info(f'{request._pyfuncitem.name}()')

    idx, restored_answer = find_most_similar(embeddings, r, a0, a1)
    logger.info(f'Restored answer a{idx} is {restored_answer}')
    pytest.assume(expected_idx == idx)


if __name__ == "__main__":
    pytest.main([__file__])






"""
A Python replacement for java.util.Properties class
This is modelled as closely as possible to the Java original.

Created - Anand B Pillai <abpillai@gmail.com>,
Update to Python 3 by Alex.
Also there are some tweeks that was done by Alex.
"""

# see https://code.activestate.com/recipes/496795-a-python-replacement-for-javautilproperties/
import sys
import re
import time
from collections import OrderedDict
import functools

class IllegalArgumentException(Exception):

    def __init__(self, lineno, msg):
        self.lineno = lineno
        self.msg = msg

    def __str__(self):
        s='Exception at line number %d => %s' % (self.lineno, self.msg)
        return s
                 
class Properties(object):
    """ A Python replacement for java.util.Properties """
    # Alex fix, handle ${}
    #curlies = re.compile('\\${.+?}')
    PLACEHOLDER_TOKEN = '$'

    def __init__(self, props=None):

        # Note: We don't take a default properties object
        # as argument yet

        # Dictionary of properties.
        # Alex changed {} to OrderedDict
        self._props = OrderedDict()
        # Dictionary of properties with 'pristine' keys
        # This is used for dumping the properties to a file
        # using the 'store' method
        # Alex changed {} to OrderedDict
        self._origprops = OrderedDict()

        # Dictionary mapping keys from property
        # dictionary to pristine dictionary
        # Alex changed {} to OrderedDict
        self._keymap = OrderedDict()
        # Alex changed {} to OrderedDict
        self._unresolved = OrderedDict()
        
        self.othercharre = re.compile(r'(?<!\\)(\s*\=)|(?<!\\)(\s*\:)')
        self.othercharre2 = re.compile(r'(\s*\=)|(\s*\:)')
        self.bspacere = re.compile(r'\\(?!\s$)')
        
    def __str__(self):
        s='{'
        for key,value in self._props.items():
            s = ''.join((s,key,'=',value,', '))

        s=''.join((s[:-2],'}'))
        return s

    def __parse(self, lines):
        """ Parse a list of lines and create
        an internal property dictionary """

        # Every line in the file must consist of either a comment
        # or a key-value pair. A key-value pair is a line consisting
        # of a key which is a combination of non-white space characters
        # The separator character between key-value pairs is a '=',
        # ':' or a whitespace character not including the newline.
        # If the '=' or ':' characters are found, in the line, even
        # keys containing whitespace chars are allowed.

        # A line with only a key according to the rules above is also
        # fine. In such case, the value is considered as the empty string.
         # In order to include characters '=' or ':' in a key or value,
        # they have to be properly escaped using the backslash character.

        # Some examples of valid key-value pairs:
        #
        # key     value
        # key=value
        # key:value
        # key     value1,value2,value3
        # key     value1,value2,value3 \
        #         value4, value5
        # key
        # This key= this value
        # key = value1 value2 value3
        
        # Any line that starts with a '#' is considerered a comment
        # and skipped. Also any trailing or preceding whitespaces
        # are removed from the key/value.
        
        # This is a line parser. It parses the
        # contents like by line.

        lineno=0
        i = iter(lines)

        for line in i:
            lineno += 1
            line = line.strip()
            # Skip null lines
            if not line: continue
            # Skip lines which are comments
            if line[0] == '#': continue
            # Some flags
            escaped=False
            # Position of first separation char
            sepidx = -1
            # A flag for performing wspace re check
            flag = 0
            # Check for valid space separation
            # First obtain the max index to which we
            # can search.
            m = self.othercharre.search(line)
            if m:
                first, last = m.span()
                start, end = 0, first
                flag = 1
                wspacere = re.compile(r'(?<![\\\=\:])(\s)')        
            else:
                if self.othercharre2.search(line):
                    # Check if either '=' or ':' is present
                    # in the line. If they are then it means
                    # they are preceded by a backslash.
                    
                    # This means, we need to modify the
                    # wspacere a bit, not to look for
                    # : or = characters.
                    wspacere = re.compile(r'(?<![\\])(\s)')        
                start, end = 0, len(line)
                
            m2 = wspacere.search(line, start, end)
            if m2:
                # print 'Space match=>',line
                # Means we need to split by space.
                first, last = m2.span()
                sepidx = first
            elif m:
                # print 'Other match=>',line
                # No matching wspace char found, need
                # to split by either '=' or ':'
                first, last = m.span()
                sepidx = last - 1
                # print line[sepidx]
                
                
            # If the last character is a backslash
            # it has to be preceded by a space in which
            # case the next line is read as part of the
            # same property
            while line[-1] == '\\':
                # Read next line
                nextline = i.next()
                nextline = nextline.strip()
                lineno += 1
                # This line will become part of the value
                line = line[:-1] + nextline

            # Now split to key,value according to separation char
            if sepidx != -1:
                key, value = line[:sepidx], line[sepidx+1:]
            else:
                key,value = line,''

            self.processPair(key, value)
            
    def processPair(self, key, value):
        """ Process a (key, value) pair """

        oldkey = key
        oldvalue = value
        
        # Create key intelligently
        keyparts = self.bspacere.split(key)
        # print keyparts

        strippable = False
        lastpart = keyparts[-1]

        if lastpart.find('\\ ') != -1:
            keyparts[-1] = lastpart.replace('\\','')

        # If no backspace is found at the end, but empty
        # space is found, strip it
        elif lastpart and lastpart[-1] == ' ':
            strippable = True

        key = ''.join(keyparts)
        if strippable:
            key = key.strip()
            oldkey = oldkey.strip()
        
        oldvalue = self.unescape(oldvalue)
        value = self.unescape(value)
        
        # Alex fix, handle ${}
        _regex ='\\${.+?}'
        if '$' != Properties.PLACEHOLDER_TOKEN:
            _regex = _regex[0:1] + Properties.PLACEHOLDER_TOKEN + _regex[2:]
        curlies = re.compile(_regex)
        found = curlies.findall(value)
        
        for f in found:
            srcKey = f[2:-1]
            if srcKey in self._props:
                value = value.replace(f, self._props[srcKey], 2)
        
        self._props[key] = value.strip()
        found = curlies.findall(value.strip())
        
        for f in found:
            srcKey = f[2:-1]
            if self._unresolved.has_key(srcKey):
                self._unresolved[srcKey] += [key]
            else:
                self._unresolved[srcKey] = [key]

        #End of Alex fix
        
        # Check if an entry exists in pristine keys
        if key in self._keymap:
            oldkey = self._keymap.get(key)
            self._origprops[oldkey] = oldvalue.strip()
        else:
            self._origprops[oldkey] = oldvalue.strip()
            # Store entry in keymap
            self._keymap[key] = oldkey
            
         # Alex fix, handle ${}
        if key in self._unresolved:
            for updateKey in self._unresolved[key]:
                self.processPair(updateKey, self._props[updateKey])
        
    def escape(self, value):

        # Java escapes the '=' and ':' in the value
        # string with backslashes in the store method.
        # So let us do the same.
        #Alex fix added r prefix to str
        newvalue = value.replace(':',r'\:')
        # Alex fix added r prefix to str
        newvalue = newvalue.replace('=',r'\=')

        return newvalue

    def unescape(self, value):

        # Reverse of escape
        # Alex fix added r prefix to str
        newvalue = value.replace(r'\:',':')
        # Alex fix added r prefix to str
        newvalue = newvalue.replace(r'\=','=')

        return newvalue    
        
    def load(self, stream):
        """ Load properties from an open file stream """

        #Alex disabled unneccesery stricked validations
        # For the time being only accept file input streams
        #if type(stream) is not file:
        #    raise TypeError('Argument should be a file object!')
        ## Check for the opened mode
        #if stream.mode != 'r':
        #    raise ValueError('Stream should be opened in read-only mode!')

        try:
            lines = stream.readlines()
            self.__parse(lines)
        except IOError as e:
            raise

    def getProperty(self, key):
        """ Return a property for the given key """
        
        return self._props.get(key,'')

    def setProperty(self, key, value):
        """ Set the property for the given key """

        if type(key) is str and type(value) is str:
            self.processPair(key, value)
        else:
            raise TypeError('both key and value should be strings!')

    def propertyNames(self):
        """ Return an iterator over all the keys of the property
        dictionary, i.e the names of the properties """

        return self._props.keys()

    def list(self, out=sys.stdout):
        """ Prints a listing of the properties to the
        stream 'out' which defaults to the standard output """

        out.write('-- listing properties --\n')
        for key,value in self._props.items():
            out.write(''.join((key,'=',value,'\n')))

    def store(self, out, header=""):
        """ Write the properties list to the stream 'out' along
        with the optional 'header' """

        if out.mode[0] != 'w':
            raise ValueError('Steam should be opened in write mode!')

        try:
            out.write(''.join(('#',header,'\n')))
            # Write timestamp
            tstamp = time.strftime('%a %b %d %H:%M:%S %Z %Y', time.localtime())
            out.write(''.join(('#',tstamp,'\n')))
            # Write properties from the pristine dictionary
            for prop, val in self._origprops.items():
                out.write(''.join((prop,'=',self.escape(val),'\n')))
                
            out.close()
        except IOError as e:
            raise

    # Alex changed name getPropertyDict() to current
    def as_dict(self):
        return self._props

    def __getitem__(self, name):
        """ To support direct dictionary like access """

        return self.getProperty(name)

    def __setitem__(self, name, value):
        """ To support direct dictionary like access """

        self.setProperty(name, value)
        
    def __getattr__(self, name):
        """ For attributes not found in self, redirect
        to the properties dictionary """

        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self._props,name):
                return getattr(self._props, name)


#https://github.com/scanny/python-pptx/blob/master/pptx/util.py
class lazyproperty(object):
    """Decorator like @property, but evaluated only on first access.

    Like @property, this can only be used to decorate methods having only
    a `self` parameter, and is accessed like an attribute on an instance,
    i.e. trailing parentheses are not used. Unlike @property, the decorated
    method is only evaluated on first access; the resulting value is cached
    and that same value returned on second and later access without
    re-evaluation of the method.

    Like @property, this class produces a *data descriptor* object, which is
    stored in the __dict__ of the *class* under the name of the decorated
    method ('fget' nominally). The cached value is stored in the __dict__ of
    the *instance* under that same name.

    Because it is a data descriptor (as opposed to a *non-data descriptor*),
    its `__get__()` method is executed on each access of the decorated
    attribute; the __dict__ item of the same name is "shadowed" by the
    descriptor.

    While this may represent a performance improvement over a property, its
    greater benefit may be its other characteristics. One common use is to
    construct collaborator objects, removing that "real work" from the
    constructor, while still only executing once. It also de-couples client
    code from any sequencing considerations; if it's accessed from more than
    one location, it's assured it will be ready whenever needed.

    Loosely based on: https://stackoverflow.com/a/6849299/1902513.

    A lazyproperty is read-only. There is no counterpart to the optional
    "setter" (or deleter) behavior of an @property. This is critically
    important to maintaining its immutability and idempotence guarantees.
    Attempting to assign to a lazyproperty raises AttributeError
    unconditionally.

    The parameter names in the methods below correspond to this usage
    example::

        class Obj(object)

            @lazyproperty
            def fget(self):
                return 'some result'

        obj = Obj()

    Not suitable for wrapping a function (as opposed to a method) because it
    is not callable.
    """

    def __init__(self, fget):
        """*fget* is the decorated method (a "getter" function).

        A lazyproperty is read-only, so there is only an *fget* function (a
        regular @property can also have an fset and fdel function). This name
        was chosen for consistency with Python's `property` class which uses
        this name for the corresponding parameter.
        """
        # ---maintain a reference to the wrapped getter method
        self._fget = fget
        # ---adopt fget's __name__, __doc__, and other attributes
        functools.update_wrapper(self, fget)

    def __get__(self, obj, type=None):
        """Called on each access of 'fget' attribute on class or instance.

        *self* is this instance of a lazyproperty descriptor "wrapping" the
        property method it decorates (`fget`, nominally).

        *obj* is the "host" object instance when the attribute is accessed
        from an object instance, e.g. `obj = Obj(); obj.fget`. *obj* is None
        when accessed on the class, e.g. `Obj.fget`.

        *type* is the class hosting the decorated getter method (`fget`) on
        both class and instance attribute access.
        """
        # ---when accessed on class, e.g. Obj.fget, just return this
        # ---descriptor instance (patched above to look like fget).
        if obj is None:
            return self

        # ---when accessed on instance, start by checking instance __dict__
        value = obj.__dict__.get(self.__name__)
        if value is None:
            # ---on first access, __dict__ item will absent. Evaluate fget()
            # ---and store that value in the (otherwise unused) host-object
            # ---__dict__ value of same name ('fget' nominally)
            value = self._fget(obj)
            obj.__dict__[self.__name__] = value
        return value

    def __set__(self, obj, value):
        """Raises unconditionally, to preserve read-only behavior.

        This decorator is intended to implement immutable (and idempotent)
        object attributes. For that reason, assignment to this property must
        be explicitly prevented.

        If this __set__ method was not present, this descriptor would become
        a *non-data descriptor*. That would be nice because the cached value
        would be accessed directly once set (__dict__ attrs have precedence
        over non-data descriptors on instance attribute lookup). The problem
        is, there would be nothing to stop assignment to the cached value,
        which would overwrite the result of `fget()` and break both the
        immutability and idempotence guarantees of this decorator.

        The performance with this __set__() method in place was roughly 0.4
        usec per access when measured on a 2.8GHz development machine; so
        quite snappy and probably not a rich target for optimization efforts.
        """
        raise AttributeError("can't set attribute")  # pragma: no cover
